'use strict'

const fs = require('fs')

// https://github.com/dominictarr/JSONStream
const JSONStream = require('JSONStream')
// https://github.com/JoshuaWise/better-sqlite3
const DBSqlite3 = require('better-sqlite3')
// https://github.com/dresende/node-sql-query
const sqlQuery = require('sql-query')
// https://github.com/lodash/lodash
const merge = require('lodash.merge')

const AllowedStatistics = {
  actualDataSize: 'actualDataSize',
  contentLength: 'contentLength',
  downloadTime: 'downloadTime',
  requestLatency: 'requestLatency',
  requestTime: 'requestTime'
}

var SQLiteFetchQueue = function (filename = 'sqliteDB.sqlite3') {
  /**
      * Path to the sqlite db file
      * @type {string}
      */
  this.sqlitePathDB = (filename === '' || filename === null) ? 'sqliteDB.sqlite3' : filename

  /**
      * The sqlite database object handle
      * @type {sqlite3.Database}
      */
  this.memorySlqliteDB = new DBSqlite3(':memory:')
}

SQLiteFetchQueue.dropDatabase = function (sqliteDatabaseName) {
  if (fs.existsSync(sqliteDatabaseName)) {
    fs.unlinkSync(sqliteDatabaseName)
  }
}

SQLiteFetchQueue.isAllowedStat = function (statisticName) {
  return Object.values(AllowedStatistics).includes(statisticName)
}

SQLiteFetchQueue.prototype.init = function () {
  const initScriptPragma = `
  -- https://www.sqlite.org/pragma.html#pragma_journal_mode
  -- You might need to increase your cache_size.
  -- Large transactions will only be speed up if the entire transaction can be stored in the cache.
  -- Memory used for the cache is allocated when it is needed, so there’s no overhead to setting it too high.
   -- A tester PRAGMA cache_size = 100000
  PRAGMA synchronous = OFF;
  PRAGMA count_changes = OFF;
  PRAGMA journal_mode = MEMORY;
  PRAGMA temp_store = MEMORY;
  -- https://www.sqlite.org/wal.html
  PRAGMA journal_mode = WAL;
  -- https://www.sqlite.org/pragma.html#pragma_wal_checkpoint
  -- PRAGMA wal_autocheckpoint = 1000;`

  const initScriptTableQueueCrawl = `
  CREATE TABLE IF NOT EXISTS QueueCrawl (
    "id" INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
    "url" TEXT NOT NULL,
    "protocol" TEXT NOT NULL,
    "host" TEXT NOT NULL,
    "port" TEXT NOT NULL,
    "path" TEXT NOT NULL,
    "uriPath" TEXT NOT NULL,
    "depth" INTEGER NOT NULL,
    "referrer" TEXT,
    "fetched" INTEGER NOT NULL,
    "status" TEXT NOT NULL,
    "stateData" TEXT NOT NULL
  );`
  const initScriptTableQueueUrls = `
  CREATE TABLE IF NOT EXISTS QueueUrls (
    "url" TEXT NOT NULL UNIQUE
  );`

  this.memorySlqliteDB.prepare(initScriptTableQueueCrawl).run()
  this.memorySlqliteDB.prepare(initScriptTableQueueUrls).run()

  this.memorySlqliteDB.prepare('ATTACH DATABASE ? AS fileDB').run(this.sqlitePathDB)
  this.memorySlqliteDB.prepare(`
  CREATE TABLE IF NOT EXISTS fileDB.QueueCrawl (
    "id" INTEGER NOT NULL,
    -- The complete, canonical URL of the resource
    "url" TEXT NOT NULL UNIQUE,
    -- The protocol of the resource (http, https)
    "protocol" TEXT NOT NULL,
    -- The full domain/hostname of the resource
    "host" TEXT NOT NULL,
    -- The port of the resource
    "port" TEXT NOT NULL,
    -- The URL path, including the query string
    "path" TEXT NOT NULL,
    -- The URL path, excluding the query string
    "uriPath" TEXT NOT NULL,
    -- How many steps simplecrawler  has taken from the initial page (which is depth 1) to this resource
    "depth" INTEGER NOT NULL,
    -- The URL of the resource where the URL of this queue item was discovered
    "referrer" TEXT,
    -- Has the request for this item been completed? You can monitor this as requests are processed
    "fetched" INTEGER NOT NULL,
    -- The internal status of the item
    "status" TEXT NOT NULL,
    -- An object containing state data and other information about the request
    "stateData" TEXT NOT NULL
  )`).run()

  this.memorySlqliteDB.exec(initScriptPragma)
  // const cache_size = this.memorySlqliteDB.pragma('synchronous', { simple: true })

  this.intervalPrune = setInterval(function (sqliteFetchQueue) {
    sqliteFetchQueue.prune()
  }, 1000 * 60 * 15, this) // prune every 15 min.
}

SQLiteFetchQueue.prototype.prune = function () {
  this.memorySlqliteDB.transaction(() => {
    if (!this.memorySlqliteDB.inTransaction) {
      throw (new Error('prune() transaction error'))
    }

    const stmtDelete = this.memorySlqliteDB.prepare('DELETE FROM QueueCrawl WHERE id = ?')
    const stmtInsert = this.memorySlqliteDB.prepare(`
    INSERT OR REPLACE INTO fileDB.QueueCrawl (id, url, protocol, host, port, path, uriPath, depth, referrer, fetched, status, stateData)
    VALUES (@id, @url, @protocol, @host, @port, @path, @uriPath, @depth, @referrer, @fetched, @status, @stateData)`)
    for (const queueItem of this.memorySlqliteDB.prepare('SELECT * FROM QueueCrawl WHERE fetched = 1').all()) {
      if (stmtInsert.run(queueItem).changes === 1) {
        stmtDelete.run(queueItem.id)
      }
    }
  })()
}

SQLiteFetchQueue.prototype.finalize = function (callback) {
  clearInterval(this.intervalPrune)

  if (this.memorySlqliteDB.open) {
    this.memorySlqliteDB.transaction(() => {
      if (!this.memorySlqliteDB.inTransaction) {
        callback(new Error('finalize() transaction error'))
        return
      }

      try {
        const stmtInsert = this.memorySlqliteDB.prepare(`
    INSERT OR REPLACE INTO fileDB.QueueCrawl (id, url, protocol, host, port, path, uriPath, depth, referrer, fetched, status, stateData)
    VALUES (@id, @url, @protocol, @host, @port, @path, @uriPath, @depth, @referrer, @fetched, @status, @stateData)`)
        for (const queueItem of this.memorySlqliteDB.prepare('SELECT * FROM QueueCrawl').all()) {
          stmtInsert.run(queueItem)
        }
      } catch (err) {
        callback(err)
      }
    })()

    this.memorySlqliteDB.prepare('DETACH fileDB').run()
    this.memorySlqliteDB.close()
  }

  callback(null)
}

/**
 * Adds an item to the queue
 * @param {QueueItem} queueItem Queue item that is to be added to the queue
 * @param {Boolean} [force=false] If true, the queue item will be added regardless of whether it already exists in the queue
 * @param {Function} callback Gets two parameters, `error` and `queueItem`. If the operation was successful, `error` will be `null` and `queueItem` will be the item that was added to the queue. It's status property will have changed to `"queued"`.
 */
SQLiteFetchQueue.prototype.add = function (queueItem, force, callback) {
  const row = this.memorySlqliteDB.prepare('SELECT * FROM QueueUrls WHERE url = ?').get(queueItem.url)
  if (row === undefined) {
    queueItem.status = 'queued'

    this.memorySlqliteDB.transaction((queueItem) => {
      if (!this.memorySlqliteDB.inTransaction) {
        callback(new Error('add() transaction error'))
      } else if (this.memorySlqliteDB.prepare('INSERT INTO QueueUrls VALUES (?)').run(queueItem.url).changes !== 1) {
        callback(new Error('add() QueueUrls'))
      } else {
        const stmtInsert = this.memorySlqliteDB
          .prepare(`
        INSERT INTO QueueCrawl (id, url, protocol, host, port, path, uriPath, depth, referrer, fetched, status, stateData)
        VALUES (?, @url, @protocol, @host, @port, @path, @uriPath, @depth, @referrer, ?, @status, ?)
        `)
          .run(null, queueItem, queueItem.fetched === true ? 1 : 0, JSON.stringify(queueItem.stateData))
        if (stmtInsert.changes !== 1) {
          callback(new Error('add() QueueCrawl'))
        } else {
          queueItem.id = stmtInsert.lastInsertRowid
          callback(null, queueItem)
        }
      }
    })(queueItem)
  } else {
    const error = new Error('Resource already exists in queue!')
    error.code = 'DUPLICATE'
    callback(error)
  }
}

/**
 * Gets the first unfetched item in the queue
 * @param {Function} callback Gets two parameters, `error` and `queueItem`. If the operation was successful, `error` will be `null`. If there are unfetched queue items left, `queueItem` will be the oldest one found. If not, `queueItem` will be `null`.
 */
SQLiteFetchQueue.prototype.oldestUnfetchedItem = function (callback) {
  const row = this.memorySlqliteDB.prepare("SELECT * FROM QueueCrawl WHERE status = 'queued' ORDER BY depth ASC LIMIT 1").get()
  if (row) { callback(null, row) } else { callback(null, null) }
}

/**
 * Updates a queue item in the queue.
 * @param {Number} id ID of the queue item that is to be updated
 * @param {Object} updates Object that will be deeply assigned (as in `Object.assign`) to the queue item. That means that nested objects will also be resursively assigned.
 * @param {Function} callback Gets two parameters, `error` and `queueItem`. If the operation was successful, `error` will be `null`.
 */
SQLiteFetchQueue.prototype.update = function (id, updates, callback) {
  const queueItem = this.memorySlqliteDB.prepare('SELECT * FROM QueueCrawl WHERE id=?').get(id)
  if (queueItem) {
    queueItem.stateData = JSON.parse(queueItem.stateData)
    // https://lodash.com/docs#merge
    merge(queueItem, updates)
    queueItem.stateData = JSON.stringify(queueItem.stateData)
    const stmtUpdate = sqlQuery.Query('sqlite')
      .update()
      .into('QueueCrawl')
      .set(updates)
      .where({ id: id })
      .build()
    if (this.memorySlqliteDB.prepare(stmtUpdate).run().changes !== 1) {
      callback(new Error('update() QueueCrawl'))
    } else {
      queueItem.stateData = JSON.parse(queueItem.stateData)
      callback(null, queueItem)
    }
  } else {
    callback(new Error('update() QueueCrawl'))
  }
}

/**
 * Counts the items in the queue that match a selector
 * @param {Object} comparator                      Comparator object used to filter items. Queue items that are counted need to match all the properties of this object.
 * @param {FetchQueue~countItemsCallback} callback
 * @param {Function} callback Gets two parameters, `error` and `items`. If the operation was successful, `error` will be `null` and `items` will be an array of QueueItems.
 */
SQLiteFetchQueue.prototype.countItems = function (comparator, callback) {
  const stmtCountItems = sqlQuery.Query('sqlite').select()
    .from('QueueCrawl')
    .count(null, 'count')
    .where(comparator)
    .build()
  const row = this.memorySlqliteDB.prepare(stmtCountItems).get()
  if (row) {
    callback(null, row.count)
  } else {
    callback(new Error('countItems() QueueCrawl'))
  }
}

/**
 * Gets the total number of queue items in the queue
 * @param {FetchQueue~getLengthCallback} callback
 * @param {Function} callback Gets two parameters, `error` and `length`. If the operation was successful, `error` will be `null`.
 */
SQLiteFetchQueue.prototype.getLength = function (callback) {
  const stmtGetLength = sqlQuery.Query('sqlite').select()
    .from('QueueCrawl')
    .count(null, 'count')
    .build()
  const row = this.memorySlqliteDB.prepare(stmtGetLength).get()
  if (row) {
    callback(null, row.count)
  } else {
    callback(new Error('getLength() QueueCrawl'))
  }
}

/**
 * Filters and returns the items in the queue that match a selector
 * @param {Object} comparator Comparator object used to filter items. Queue items that are returned need to match all the properties of this object.
 * @param {Function} callback Gets two parameters, `error` and `items`. If the operation was successful, `error` will be `null` and `items` will be an array of QueueItems.
 */
SQLiteFetchQueue.prototype.filterItems = function (comparator, callback) {
  const stmtFilterItems = sqlQuery.Query('sqlite').select()
    .from('QueueCrawl')
    .where(comparator)
    .build()

  try {
    callback(null, this.memorySlqliteDB.prepare(stmtFilterItems).all())
  } catch (error) {
    callback(error)
  }
}

/**
 * Get a queue item by index
 * @param {Number} index The index of the queue item in the queue
 * @param {Function} callback Gets two parameters, `error` and `queueItem`. If the operation was successful, `error` will be `null`.
 */
SQLiteFetchQueue.prototype.get = function (index, callback) {
  const stmtSelect = sqlQuery.Query('sqlite').select()
    .from('QueueCrawl')
    .where({ id: index })
    .build()

  const row = this.memorySlqliteDB.prepare(stmtSelect).get()
  if (row) {
    row.stateData = JSON.parse(row.stateData)
    callback(null, row)
  } else {
    callback(new Error('Index was not found'))
  }
}

/**
 * Gets the maximum value of a stateData property from all the items in the
 * queue. This means you can eg. get the maximum request time, download size
 * etc.
 * @param {String} statisticName Can be any of the strings in {@link FetchQueue._allowedStatistics}
 * @param {Function} callback Gets two parameters, `error` and `max`. If the operation was successful, `error` will be `null`.
 */
SQLiteFetchQueue.prototype.max = function (statisticName, callback) {
  try {
    if (!Object.values(AllowedStatistics).includes(statisticName)) { return callback(new Error('Invalid statistic')) }
    callback(null, this.memorySlqliteDB.prepare(`SELECT MAX(json_extract(stateData, '$.${statisticName}')) AS max FROM QueueCrawl`).get().max)
  } catch (error) {
    callback(error)
  }
}

/**
 * Gets the minimum value of a stateData property from all the items in the
 * queue. This means you can eg. get the minimum request time, download size
 * etc.
 * @param {String} statisticName Can be any of the strings in {@link FetchQueue._allowedStatistics}
 * @param {Function} callback Gets two parameters, `error` and `min`. If the operation was successful, `error` will be `null`.

 */
SQLiteFetchQueue.prototype.min = function (statisticName, callback) {
  try {
    if (!Object.values(AllowedStatistics).includes(statisticName)) { return callback(new Error('Invalid statistic')) }
    callback(null, this.memorySlqliteDB.prepare(`SELECT MIN(json_extract(stateData, '$.${statisticName}')) AS min FROM QueueCrawl`).get().min)
  } catch (error) {
    callback(error)
  }
}

/**
 * Gets the average value of a stateData property from all the items in the
 * queue. This means you can eg. get the average request time, download size
 * etc.
 * @param {String} statisticName Can be any of the strings in {@link FetchQueue._allowedStatistics}
 * @param {Function} callback Gets two parameters, `error` and `avg`. If the operation was successful, `error` will be `null`.
 */
SQLiteFetchQueue.prototype.avg = function (statisticName, callback) {
  try {
    if (!Object.values(AllowedStatistics).includes(statisticName)) { return callback(new Error('Invalid statistic')) }
    callback(null, this.memorySlqliteDB.prepare(`SELECT AVG(json_extract(stateData, '$.${statisticName}')) AS avg FROM QueueCrawl`).get().avg)
  } catch (error) {
    callback(error)
  }
}

/**
 * Checks if a URL already exists in the queue. Returns the number of occurences
 * of that URL.
 * @param {String} url URL to check the existence of in the queue
 * @param {Function} callback Gets two parameters, `error` and `count`. If the operation was successful, `error` will be `null`.
 */
SQLiteFetchQueue.prototype.exists = function (url, callback) {
  try {
    const row = this.memorySlqliteDB.prepare('SELECT * FROM QueueCrawl WHERE url = ? LIMIT 1').get(url)
    callback(null, row ? 1 : 0)
  } catch (error) {
    callback(error, 0)
  }
}

SQLiteFetchQueue.prototype.dropQueue = function (callback) {
  try {
    this.memorySlqliteDB.prepare('DROP TABLE IF EXISTS QueueCrawl').run()
    this.memorySlqliteDB.prepare('VACUUM').run()
    callback(null)
  } catch (error) {
    callback(error)
  }
}

/**
 * Writes the queue to disk in a JSON file. This file can later be imported
 * using {@link FetchQueue#defrost}
 * @param {String} filename Filename passed directly to [fs.writeFile]{@link https://nodejs.org/api/fs.html#fs_fs_writefile_file_data_options_callback}
 * @param {Function} callback Gets a single `error` parameter. If the operation was successful, this parameter will be `null`.
 */
SQLiteFetchQueue.prototype.freeze = function (filename, callback) {
  try {
    if (fs.existsSync(filename)) { fs.unlinkSync(filename) }

    // Pipe the JSON data to the file.
    const jsonwriter = JSONStream.stringify()
    jsonwriter.pipe(fs.createWriteStream(filename))

    this.memorySlqliteDB.transaction(() => {
      if (!this.memorySlqliteDB.inTransaction) {
        callback(new Error('freeze() transaction error'))
        return
      }

      const stmtInsert = this.memorySlqliteDB.prepare(`
      INSERT OR REPLACE INTO fileDB.QueueCrawl (id, url, protocol, host, port, path, uriPath, depth, referrer, fetched, status, stateData)
      VALUES (@id, @url, @protocol, @host, @port, @path, @uriPath, @depth, @referrer, @fetched, @status, @stateData)`)
      for (const queueItem of this.memorySlqliteDB.prepare('SELECT * FROM QueueCrawl').all()) {
        stmtInsert.run(queueItem)
      }
    })()

    const stmtSelectAll = this.memorySlqliteDB.prepare('SELECT * FROM fileDB.QueueCrawl')
    for (const row of stmtSelectAll.iterate()) { jsonwriter.write(row) }

    jsonwriter.end()
    callback(null)
  } catch (error) {
    callback(error)
  }
}

/**
 * Import the queue from a frozen JSON file on disk.
 * @param {String} filename Filename passed directly to [fs.readFile]{@link https://nodejs.org/api/fs.html#fs_fs_readfile_file_options_callback}
 * @param {Function} callback Gets a single `error` parameter. If the operation was successful, this parameter will be `null`.
 */
SQLiteFetchQueue.prototype.defrost = function (filename, callback) {
  fs.createReadStream(filename)
    .pipe(JSONStream.parse('*'))
    .on('error', error => {
      callback(error)
    })
    .on('data', data => {
      this.add(data, false, err => { if (err) { callback(err) } })
    })
    .on('end', () => {
      this.getLength((err, rowCount) => {
        if (err) { callback(err) } else { callback(null, rowCount) }
      })
    })
}

module.exports = SQLiteFetchQueue
