'use strict'

const querystring = require('querystring')

// https://github.com/lorenwest/node-config
const config = require('config')
// https://github.com/simplecrawler/simplecrawler
const Crawler = require('simplecrawler')

const SQLiteFetchQueue = require('../SQLiteFetchQueue.js')

function getDate () {
  const currentdate = new Date()
  return (
    (currentdate.getDate() < 10 ? '0' : '') + currentdate.getDate() +
    '/' +
    (currentdate.getMonth() + 1 < 10 ? '0' : '') + (currentdate.getMonth() + 1) +
    '/' +
    currentdate.getFullYear()
  )
}

function getTime () {
  const currentdate = new Date()
  return (
    (currentdate.getHours() < 10 ? '0' : '') + currentdate.getHours() +
    ':' +
    (currentdate.getMinutes() < 10 ? '0' : '') + currentdate.getMinutes() +
    ':' +
    (currentdate.getSeconds() < 10 ? '0' : '') + currentdate.getSeconds()
  )
}

const siteToCrawl = process.argv[2]

var sqliteDatabaseName = ''
try {
  const hostname = new URL(siteToCrawl).hostname
  sqliteDatabaseName = `${hostname}.sqlite3`
} catch (e) {
  console.error(`Invalid Url ${siteToCrawl} - ${e}`)
  process.exit(1)
}

const crawler = new Crawler(querystring.unescape(siteToCrawl))
// Crawler's user agent string
crawler.userAgent = config.has('Politeness.UserAgent') ? config.get('Politeness.UserAgent') : 'SimpleCrawler'
//  interval at which new requests are spawned by the crawler
crawler.interval = 250
// Controls the max depth of resources that the crawler fetches
crawler.maxDepth = config.has('CrawlBehavior.LimitSearchDepth') ? config.get('CrawlBehavior.LimitSearchDepth') : 2
// Controls whether the crawler is allowed to change the host setting if the first response is a redirect to another domain.
crawler.allowInitialDomainChange = false
// Controls whether the crawler fetches only URL's where the hostname matches host
crawler.filterByDomain = true
// Don't can for new resources inside of HTML comments
crawler.parseHTMLComments = false
// Don't scan for new resources inside of <script> tags
crawler.parseScriptTags = false
// Controls whether the crawler respects the robots.txt rules of any domain.
crawler.respectRobotsTxt = true
// Maximum time we'll wait for headers
crawler.timeout = 17000
// Maximum request concurrency
crawler.maxConcurrency = 10
// Don't fetch URL's that points to a subdomain of host
crawler.scanSubdomains = false
// HTTP proxy should be used for requests
crawler.useProxy = config.has('Politeness.UseProxy') ? config.get('Politeness.UseProxy') : false
// Hostname to use for the proxy
crawler.proxyHostname = config.has('Politeness.ProxyHostname') ? config.get('Politeness.ProxyHostname') : ''
// Port to use for the proxy
crawler.proxyPort = config.has('Politeness.ProxyPort') ? config.get('Politeness.ProxyPort') : 0
// Page number
crawler.pageIndex = 0

crawler.on('fetchclienterror', function (queueItem, error) {
  console.error(`${getDate()} ${getTime()} ERROR Client request: ${error}`)
  process.exit(1)
})
crawler.on('fetch404', function (queueItem, response) {
  crawler.emit('fetcherror', queueItem, response)
})
crawler.on('fetch410', function (queueItem, response) {
  crawler.emit('fetcherror', queueItem, response)
})
crawler.on('fetchtimeout', function (queueItem, timeout) {
  queueItem.pageIndex = crawler.pageIndex++
  console.log(`${getDate()} ${getTime()} [${queueItem.pageIndex}-${queueItem.depth}] Time Out (${timeout}) ${queueItem.url})`)
})
crawler.on('fetcherror', function (queueItem, response) {
  queueItem.pageIndex = crawler.pageIndex++
  console.log(`${getDate()} ${getTime()} [${queueItem.pageIndex}-${queueItem.depth}] ERROR ${queueItem.url} - ${response.statusCode} (${response.statusMessage})`)
})
crawler.on('fetchredirect', function (queueItem, redirectQueueItem, response) {
  queueItem.pageIndex = crawler.pageIndex++
  console.log(`${getDate()} ${getTime()} [${queueItem.pageIndex}-${queueItem.depth}] ${queueItem.url} redirected to ${redirectQueueItem.url}`)
})
crawler.on('fetchcomplete', function (queueItem, responseBuffer, response) {
  if (queueItem.status === 'downloaded') {
    queueItem.pageIndex = crawler.pageIndex++
    console.log(`${getDate()} ${getTime()} [${queueItem.pageIndex}-${queueItem.depth}] ${querystring.unescape(queueItem.url)} (${responseBuffer.length} bytes)`)
  }
})
crawler.on('complete', function () {
  console.log(`${getDate()} ${getTime()} Saving data to SQLite database: '${this.queue.sqlitePathDB}'`)
  this.queue.finalize(function (err) {
    if (err) {
      console.error(err)
      process.exit(1)
    }
    console.log(`${getDate()} ${getTime()} Crawl is successfully finished!`)
    process.exit(0)
  })
})

crawler.addFetchCondition(function (queueItem, referrerQueueItem, callback) {
  if (referrerQueueItem.status === 'redirected' && queueItem.depth > crawler.maxDepth) { callback(null, false) } else { callback(null, true) }
})
crawler.addDownloadCondition(function (queueItem, response, callback) {
  const mimeTypeSupported = [/^text\/html/i].some(function (mimeCheck) {
    return mimeCheck.test(queueItem.stateData.contentType)
  })
  callback(null, mimeTypeSupported)
})

process.on('exit', () => { crawler.queue.finalize((err) => { if (err) { console.error(err) } }) })
process.on('SIGHUP', () => process.exit(128 + 1))
process.on('SIGINT', () => process.exit(128 + 2))
process.on('SIGTERM', () => process.exit(128 + 15))

try {
  const crawlerQueue = new SQLiteFetchQueue(sqliteDatabaseName)
  crawlerQueue.init()
  crawler.queue = crawlerQueue

  console.log(`${getDate()} ${getTime()} Starting crawl on ${crawler.initialURL}`)

  crawler.start()
} catch (err) {
  console.error(err)
  process.exit(1)
}
