/* eslint-disable no-unused-expressions */
/* eslint-env mocha */

'use strict'

const path = require('path')

// https://github.com/simplecrawler/simplecrawler
const Crawler = require('simplecrawler')
// https://github.com/chaijs/chai
const chai = require('chai')
const should = chai.should()

const SQLiteFetchQueue = require('../SQLiteFetchQueue.js')

const queueTestData = require('./queueTestData.json')

describe('SQLiteFetchQueue methods', function () {
  const siteToCrawl = 'http://www.test.com/'
  const sqliteDatabaseName = `${path.join(__dirname, new URL(siteToCrawl).hostname)}.sqlite3`

  const crawler = new Crawler(siteToCrawl)
  const crawlerQueue = new SQLiteFetchQueue(sqliteDatabaseName, 'test_queue_crawl')

  after(function (done) {
    crawler.queue.finalize(function (err) {
      should.not.exist(err)
      done(err)
    })
  })

  it('should init the queue', function (done) {
    try {
      crawlerQueue.init()
      crawler.queue = crawlerQueue
      done()
    } catch (err) {
      done(err)
    }
  })

  it('should add one row to the queue', function (done) {
    crawler.queue.add(queueTestData[0], false, function (err, queueItem) {
      should.not.exist(err)
      crawler.queue.getLength(function (error, length) {
        length.should.equal(1)
        done(error)
      })
    })
  })

  it('should add three rows to the queue', function (done) {
    crawler.queue.add(queueTestData[1], false, function (err, queueItem) {
      should.not.exist(err)
      crawler.queue.add(queueTestData[2], false, function (err, queueItem) {
        should.not.exist(err)
        crawler.queue.add(queueTestData[3], false, function (err, queueItem) {
          should.not.exist(err)
          crawler.queue.getLength(function (error, length) {
            length.should.equal(4)
            done(error)
          })
        })
      })
    })
  })

  it("shouldn't add duplicate to the queue", function (done) {
    crawler.queue.add(queueTestData[0], false, function (err, queueItem) {
      err.should.be.an('error')
      done()
    })
  })

  it('should update items in the queue', function (done) {
    crawler.queue.update(queueTestData[0].id, {
      status: 'downloaded',
      fetched: true
    }, function (err, queueItem) {
      should.not.exist(err)
      queueItem.should.include({
        url: queueTestData[0].url,
        status: 'downloaded',
        fetched: true
      })

      crawler.queue.update(queueTestData[1].id, {
        status: 'notfound',
        fetched: true
      }, function (err, queueItem) {
        should.not.exist(err)
        queueItem.should.include({
          url: queueTestData[1].url,
          status: 'notfound',
          fetched: true
        })

        crawler.queue.update(queueTestData[2].id, {
          status: 'downloaded',
          fetched: true
        }, function (err, queueItem) {
          should.not.exist(err)
          queueItem.should.include({
            url: queueTestData[2].url,
            status: 'downloaded',
            fetched: true
          })

          done()
        })
      })
    })
  })

  it('should get items from the queue', function (done) {
    crawler.queue.get(3, function (err, item) {
      should.not.exist(err)
      item.url.should.equal(queueTestData[2].url)
      done()
    })
  })

  it('should error when getting queue items out of range', function (done) {
    crawler.queue.getLength(function (err, length) {
      should.not.exist(err)
      crawler.queue.get(length * 2, function (err, queueItem) {
        should.not.exist(queueItem)
        err.should.be.an('error')
        done()
      })
    })
  })

  it('should get the oldest unfetched item', function (done) {
    crawler.queue.oldestUnfetchedItem(function (err, queueItem) {
      should.not.exist(err)
      queueItem.url.should.equal('http://www.test.com/stage/3')
      done()
    })
  })

  it('should get a max statistic from the queue', function (done) {
    crawler.queue.max('downloadTime', function (err, max) {
      should.not.exist(err)
      max.should.be.a('number')
      max.should.equal(2)
      done()
    })
  })

  it('should get a min statistic from the queue', function (done) {
    crawler.queue.min('requestTime', function (err, min) {
      should.not.exist(err)
      min.should.be.a('number')
      min.should.equal(2)
      done()
    })
  })

  it('should verify against allowed statistics', function () {
    chai.expect(SQLiteFetchQueue.isAllowedStat('actualDataSize')).to.be.true
    chai.expect(SQLiteFetchQueue.isAllowedStat('contentLength')).to.be.ok
    chai.expect(SQLiteFetchQueue.isAllowedStat('downloadTime')).to.be.ok
    chai.expect(SQLiteFetchQueue.isAllowedStat('requestLatency')).to.be.ok
    chai.expect(SQLiteFetchQueue.isAllowedStat('requestTime')).to.be.ok
    chai.expect(SQLiteFetchQueue.isAllowedStat('isNotAllowed')).to.be.false
  })

  it('should get an average statistic from the queue', function (done) {
    crawler.queue.avg('contentLength', function (err, avg) {
      should.not.exist(err)
      avg.should.be.a('number')
      avg.should.equal((68 + 14 + 37) / 3)
      done()
    })
  })

  it("shouldn't get an invalid statistic from the queue", function (done) {
    crawler.queue.max('invalidMax', function (err, avg) {
      err.should.be.an('error')
      crawler.queue.max('invalidMin', function (err, avg) {
        err.should.be.an('error')
        crawler.queue.max('invalidAvg', function (err, avg) {
          err.should.be.an('error')
          done()
        })
      })
    })
  })

  it('should get the number of completed queue items', function (done) {
    crawler.queue.countItems({ fetched: true }, function (err, complete) {
      should.not.exist(err)
      complete.should.be.a('number')
      complete.should.equal(3)
      done(err)
    })
  })

  it('should get queue items with a specific status', function (done) {
    crawler.queue.filterItems({ depth: 2 }, function (err, items) {
      should.not.exist(err)
      items.should.be.an('array')
      items.map(function (item) {
        return item.url
      }).should.eql(['http://www.test.com/404', 'http://www.test.com/stage/2'])
      done(err)
    })
  })

  it('should item exists in the queue', function (done) {
    crawler.queue.exists('http://www.test.com/', function (err, result) {
      should.not.exist(err)
      result.should.be.a('number')
      result.should.equal(1)
      done(err)
    })
  })

  it("shouldn't item exists in the queue", function (done) {
    crawler.queue.exists('http://www.example.com/', function (err, result) {
      should.not.exist(err)
      result.should.be.a('number')
      result.should.equal(0)
      done(err)
    })
  })

  it('should count items with a specific status', function (done) {
    crawler.queue.countItems({ status: 'downloaded' }, function (err, count) {
      should.not.exist(err)
      count.should.be.a('number')
      count.should.equal(2)
      done(err)
    })
  })

  it('should get items that have failed', function (done) {
    crawler.queue.countItems({ status: 'failed' }, function (err, count) {
      should.not.exist(err)
      count.should.be.a('number')
      count.should.equal(0)

      crawler.queue.countItems({ status: 'notfound' }, function (err, count) {
        should.not.exist(err)
        count.should.be.a('number')
        count.should.equal(1)
        done(err)
      })
    })
  })

  it('should count items with a 200 HTTP status', function (done) {
    crawler.queue.countItems({
      __sql: [["json_extract(stateData, '$.code') = 200"]]
    }, function (err, count) {
      should.not.exist(err)
      count.should.be.a('number')
      count.should.equal(2)
      done(err)
    })
  })

  it('should save database to JSON file', function (done) {
    crawler.queue.freeze(`${sqliteDatabaseName}.json`, function (err, result) {
      should.not.exist(err)
      done(err)
    })
  })

  it('should restore database from JSON file', function (done) {
    crawler.queue.dropQueue(function (err) {
      should.not.exist(err)
      chai.expect(() => (crawler.queue.memorySlqliteDB.prepare('DROP TABLE IF EXISTS QueueUrls').run())).not.to.throw(TypeError)
      chai.expect(() => (crawler.queue.init())).not.to.throw(TypeError)
      crawler.queue.getLength(function (err, length) {
        should.not.exist(err)
        length.should.be.a('number')
        length.should.equal(0)
      })
      crawler.queue.defrost(`${sqliteDatabaseName}.json`, function (err, result) {
        should.not.exist(err)
        result.should.be.a('number')
        result.should.equal(4)
        done(err)
      })
    })
  })
})
